/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * JBIDescriptorBuilder - Initial Developper : ofabre - EBM Websourcing
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.descriptor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.jbi.descriptor.util.InputStreamWrapper;
import org.objectweb.petals.jbi.descriptor.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * It's the Petals JBIDescriptor builder.
 *
 * @author ofabre - EBM Websourcing
 *
 */
public final class JBIDescriptorBuilder {

    /**
     * JBI Schema
     */
    public static final String JBI_XSD = "jbi.xsd";

    /**
     * PEtALS Extensions
     */
    public static final String PETALS_EXTENSIONS_XSD = "extensions.xsd";

    /**
     * PEtALS Schema (conform to JBI specification)
     */
    public static final String PETALS_XSD = "petalsJBI.xsd";

    /**
     * JBI NAMESPACE
     */
    public static final String NAMESPACE_JBI = "http://java.sun.com/xml/ns/jbi";

    /**
     * XSI SCHEMA
     */
    public static final String NAMESPACE_XSI = "http://www.w3.org/2001/XMLSchema-instance";

    /**
     * XSD SCHEMA
     */
    public static final String NAMESPACE_XSD = "http://www.w3.org/2001/XMLSchema";

    /**
     * Standard location to the JBI Descriptor resource.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF" + File.separator + "jbi.xml";

    /**
     * SCHEMA RESOURCE
     */
    public static final String SCHEMA_RESOURCE = "/schema/jbi";

    /**
     * The JBI Descriptor XML Schema Resource Location ("schema/jbi.xsd").
     */
    public static final String JBI_SCHEMA_RESOURCE = JBIDescriptorBuilder.SCHEMA_RESOURCE
            + File.separator + JBIDescriptorBuilder.JBI_XSD;

    private static final String BOOTSTRAP_CLASS_NAME_ELEM_NAME = "bootstrap-class-name";

    private static final String BOOTSTRAP_CLASS_PATH_ELEM_NAME = "bootstrap-class-path";

    private static final String COMPONENT_CLASS_NAME_ELEM_NAME = "component-class-name";

    private static final String COMPONENT_CLASS_PATH_ELEM_NAME = "component-class-path";

    private static final String ENDPOINT_ATTR_NAME = "endpoint-name";

    private static final String IDENTIFICATION_ELEM_NAME = "identification";

    private static final String INTERFACE_ATTR_NAME = "interface-name";

    private static final String SERVICE_ATTR_NAME = "service-name";

    private static final String SHARED_LIBRARY_ELEM_NAME = "shared-library";

    private static final String VERSION_ATTR_NAME = "version";

    private static final String PETALS_HOME_PROPERTY = "petals.home";

    /**
     * Singleton instance
     */
    private static JBIDescriptorBuilder ref;

    private JBIDescriptorBuilder() {
        // nothing to do
    }

    /**
     *
     * @return a single instance of the {@link JBIDescriptorBuilder}
     * @throws JBIDescriptorException
     *             if it can't correctly load jbi.xsd schema file
     */
    public static JBIDescriptorBuilder getInstance() throws JBIDescriptorException {
        if (JBIDescriptorBuilder.ref == null) {
            JBIDescriptorBuilder.ref = new JBIDescriptorBuilder();

        }
        return JBIDescriptorBuilder.ref;
    }

    /**
     * Extension nodes are nodes which names are different from those specified
     * by JBI. This method load these extensions.
     *
     * @param node
     *            an extension xml node
     * @throws JBIDescriptorException
     */
    private void loadExtensions(final Extensions extensions, final Node node)
            throws JBIDescriptorException {

        // Add the extension node to the extensions documentfragment containing
        // all extensions
        DocumentFragment fragment = extensions.getDocumentFragment();
        if (fragment == null) {
            fragment = node.getOwnerDocument().createDocumentFragment();
            extensions.setDocumentFragment(fragment);
        }
        fragment.appendChild(node);
    }

    /**
     * Returns a String representation of a JBI descriptor file
     *
     * @param jbiXmlUri
     *            the JBI Descriptor file <code>URI</code>. Must not be null
     *            and must point on a real jbi.xml file
     * @return a <code>String</code> representation of the JBI Descriptor
     * @throws JBIDescriptorException
     */
    public String getJbiDescriptorAsString(final URI jbiXmlUri) throws JBIDescriptorException {
        if (jbiXmlUri == null) {
            throw new IllegalArgumentException("jbiXmlUri must not be null");
        }
        final File descriptorFile = new File(jbiXmlUri);
        InputStream is = null;
        if (descriptorFile.isFile()) {
            try {
                is = new FileInputStream(descriptorFile);
            } catch (final FileNotFoundException e) {
                throw new JBIDescriptorException("Can't find jbi descriptor file: "
                        + jbiXmlUri.toString(), e);
            }
        } else {
            throw new IllegalArgumentException("Given URI must point on a real jbi.xml file :"
                    + jbiXmlUri);
        }
        return this.getJbiDescriptorAsString(is);
    }

    /**
     * Returns a String representation of a JBI descriptor file
     *
     * @param jbiXmlUri
     *            the JBI Descriptor file <code>URI</code>. Must not be null.
     * @return a <code>String</code> representation of the JBI Descriptor
     * @throws JBIDescriptorException
     */
    public String getJbiDescriptorAsString(final InputStream jbiXmlStream)
            throws JBIDescriptorException {
        if (jbiXmlStream == null) {
            throw new IllegalArgumentException("jbiXmlStream must not be null");
        }
        String ret = null;
        ByteArrayOutputStream os = null;
        try {
            os = new ByteArrayOutputStream();
            IOUtils.copy(jbiXmlStream, os);
            ret = os.toString();
        } catch (final IOException e) {
            throw new JBIDescriptorException("Can't read descriptor as String", e);
        }
        return ret;
    }

    /**
     * Validate the jbi.xml file denoted by the given {@link URI} and create a
     * {@link JBIDescriptor} from it
     *
     * @param jbiXmlUri
     * @return {@link JBIDescriptor} representation of the jbi.xml file
     * @throws JBIDescriptorException
     */
    public JBIDescriptor validateAndBuildDescriptor(final URI jbiXmlUri)
            throws JBIDescriptorException {
        this.validateJbiXml(jbiXmlUri);
        final Document document = this.parseJbiXml(jbiXmlUri);
        return this.createJBIDescriptor(document);
    }

    /**
     * Validate the jbi.xml file contained in the {@link InputStream} and create
     * a {@link JBIDescriptor} from it
     *
     * @param jbiXmlStream
     * @return {@link JBIDescriptor} representation of the jbi.xml file
     * @throws JBIDescriptorException
     */
    public JBIDescriptor validateAndBuildDescriptor(final InputStream jbiXmlStream)
            throws JBIDescriptorException {
        // Wrap the inputStream to protect the Stream from the closings of the
        // parser.

        final InputStreamWrapper inputStream = new InputStreamWrapper(jbiXmlStream);
        this.validateJbiXml(inputStream);
        final Document document = this.parseJbiXml(inputStream);

        return this.createJBIDescriptor(document);
    }

    /**
     * Create a {@link JBIDescriptor} from jbi.xml file without validation
     *
     * @param jbiXmlUri
     * @return {@link JBIDescriptor} representation of the jbi.xml file
     * @throws JBIDescriptorException
     */
    public JBIDescriptor buildDescriptor(final URI jbiXmlUri) throws JBIDescriptorException {
        final Document document = this.parseJbiXml(jbiXmlUri);
        return this.createJBIDescriptor(document);
    }

    /**
     * Create a {@link JBIDescriptor} from jbi.xml file without validation
     *
     * @param jbiXmlStream
     * @return {@link JBIDescriptor} representation of the jbi.xml file
     * @throws JBIDescriptorException
     */
    public JBIDescriptor buildDescriptor(final InputStream jbiXmlStream)
            throws JBIDescriptorException {
        final Document document = this.parseJbiXml(jbiXmlStream);
        return this.createJBIDescriptor(document);
    }

    /**
     *
     * @param sourceURI
     *            {@link URI} of a jbi.xml descriptor file
     * @throws JBIDescriptorException
     *             if an error occurs during jbi.xml validation
     */
    public void validateJbiXml(final URI sourceURI) throws JBIDescriptorException {
        this.validateJbiXml(new StreamSource(new File(sourceURI)));
    }

    /**
     *
     * @param sourceStream
     *            {@link InputStream} containing jbi.xml descriptor file
     * @throws JBIDescriptorException
     *             if an error occurs during jbi.xml validation
     */
    public void validateJbiXml(final InputStream sourceStream) throws JBIDescriptorException {
        this.validateJbiXml(new StreamSource(sourceStream));
    }

    /**
     *  Realize the PEtALS - JBI validation. This method tries to realize the validation
     *  with the PEtALS schema. If the validation fails, the validation is realize by the JBI
     *  schema.
     * @param source
     *            a {@link Source} containing jbi.xml descriptor file stream
     * @throws JBIDescriptorException
     *             if an error occurs during jbi.xml validation
     */
    public void validateJbiXml(final Source source) throws JBIDescriptorException {
        // create a Validator instance which is be used to validate a
        // document

        try {
            // Realize the petals validation
            this.petalsValidation(source);
        } catch (final JBIDescriptorException e) {
            // If the Petals validation fails, start the jbi validation
            this.jbiValidation(source);
        }
    }

    /**
     * Realize the PEtALS validation (Verify if the document is conform to PEtALS requirement)
     * @param source the source
     * @throws JBIDescriptorException
     *      the source is not conform
     */
    public void petalsValidation(final Source source) throws JBIDescriptorException {
        // create a Validator instance which is be used to validate a
        // document

        final Schema petalsSchema = this.getPetalsSchema();

        final Validator validator = petalsSchema.newValidator();

        // validate the DOM tree
        try {
            validator.validate(source);
        } catch (final SAXException e) {
            throw new JBIDescriptorException("Petals Validation: Error validating jbi.xml file", e);
        } catch (final IOException e) {
            throw new JBIDescriptorException("Petals Validation: Error validating jbi.xml file", e);
        }
    }

    /**
     * Realize the JBI validation (Verify if the document is conform to JBI requirement)
     * @param source the source
     * @throws JBIDescriptorException
     *      the source is not conform
     */
    public void jbiValidation(final Source source) throws JBIDescriptorException {
        // create a Validator instance which is be used to validate a
        // document

        final Schema jbiSchema = this.getJBISchema();

        final Validator validator = jbiSchema.newValidator();

        // validate the DOM tree
        try {
            validator.validate(source);
        } catch (final SAXException e) {
            throw new JBIDescriptorException("JBI Validation (SAXException): "
                    + "Error validating jbi.xml file", e);
        } catch (final IOException e) {
            throw new JBIDescriptorException("JBI Validation (IOException): "
                    + "Error validating jbi.xml file", e);
        }
    }

    /**
     * Parse a jbi.xml file to create a {@link Document} representation
     *
     * @param sourceURI
     *            an {@link URI} pointing on a jbi.xml file
     * @return a {@link Document} representation of the jbi.xml file
     * @throws JBIDescriptorException
     *             if it can't create DocumentBuilder or if an error occurs
     *             during file parsing
     */
    public Document parseJbiXml(final URI sourceURI) throws JBIDescriptorException {
        try {
            return this.parseJbiXml(new FileInputStream(new File(sourceURI)));
        } catch (final FileNotFoundException e) {
            throw new JBIDescriptorException("Can't find jbi.xml file", e);
        }
    }

    /**
     * Parse a jbi.xml file to create a {@link Document} representation
     *
     * @param source
     *            an {@link InputStream} containing the jbi.xml file stream
     * @return a {@link Document} representation of the jbi.xml file
     * @throws JBIDescriptorException
     *             if it can't create DocumentBuilder or if an error occurs
     *             during file parsing
     */
    public Document parseJbiXml(final InputStream source) throws JBIDescriptorException {

        final DocumentBuilder docBuilder = this.createDocumentBuilder();

        Document document = null;

        try {
            document = docBuilder.parse(source);
        } catch (final SAXException saxe) {
            throw new JBIDescriptorException("JBI descriptor is not well formed XML", saxe);
        } catch (final IOException ioe) {
            throw new JBIDescriptorException("Can't read JBI descriptor", ioe);
        }
        return document;
    }

    /**
     * create the document builder.
     * @return
     * @throws JBIDescriptorException
     */
    private DocumentBuilder createDocumentBuilder() throws JBIDescriptorException {
        final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        docBuilderFactory.setNamespaceAware(true);

        DocumentBuilder docBuilder = null;

        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
        } catch (final ParserConfigurationException pce) {
            throw new JBIDescriptorException("Bad XML parser configuration", pce);
        }
        return docBuilder;
    }

    /**
     * Parse the XML document and load values to the Descriptor object
     * representation.
     *
     * @param document
     * @return a populated Descriptor instance with values loaded from the XML
     *         file. May return <code>null</code> if an error occur while
     *         navigating on XML tree.
     * @throws JBIDescriptorException
     */
    public JBIDescriptor createJBIDescriptor(final Document document) throws JBIDescriptorException {

        final JBIDescriptor descriptor = new JBIDescriptor();

        // get root element
        final Element jbi = document.getDocumentElement();

        // extract JBI version attribute from the root element
        String versionAttr;
        try {
            versionAttr = XMLUtil.getRequiredAttributeValue(jbi,
                    JBIDescriptorBuilder.VERSION_ATTR_NAME);
        } catch (final NullPointerException e) {
            throw new JBIDescriptorException(e);
        }

        descriptor.setVersion(Double.parseDouble(versionAttr));

        final List<Node> children = XMLUtil.getNodeChildren(jbi);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if ("component".equals(child.getLocalName())) {
                    descriptor.setComponent(this.loadComponentXml(child));
                } else if (JBIDescriptorBuilder.SHARED_LIBRARY_ELEM_NAME.equals(child
                        .getLocalName())) {
                    descriptor.setSharedLibrary(this.loadSharedLibraryXml(child));
                } else if ("service-assembly".equals(child.getLocalName())) {
                    descriptor.setServiceAssembly(this.loadServiceAssemblyXml(child));
                } else if ("services".equals(child.getLocalName())) {
                    descriptor.setServices(this.loadServicesXml(child));
                }
            }
        }
        return descriptor;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     * @throws
     */
    private ComponentDescription loadComponentXml(final Node production)
            throws JBIDescriptorException {
        final ComponentDescription component = new ComponentDescription();

        // component attributes
        try {
            component.setType(XMLUtil.getRequiredAttributeValue(production, "type"));
        } catch (final NullPointerException e) {
            throw new JBIDescriptorException(e);
        }
        component.setComponentClassLoaderDelegation(XMLUtil.getAttributeValue(production,
                "component-class-loader-delegation"));
        component.setBootstrapClassLoaderDelegation(XMLUtil.getAttributeValue(production,
                "bootstrap-class-loader-delegation"));

        final Extensions extensions = new Extensions();
        final List<SharedLibraryList> slList = new ArrayList<SharedLibraryList>();
        // component sub nodes
        final List<Node> children = XMLUtil.getNodeChildren(production);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if (JBIDescriptorBuilder.IDENTIFICATION_ELEM_NAME.equals(child.getLocalName())) {
                    component.setIdentification(this.loadIdentificationXml(child));
                } else if (JBIDescriptorBuilder.COMPONENT_CLASS_NAME_ELEM_NAME.equals(child
                        .getLocalName())) {
                    component.setComponentClassName(XMLUtil.getTextContent(child));
                } else if (JBIDescriptorBuilder.COMPONENT_CLASS_PATH_ELEM_NAME.equals(child
                        .getLocalName())) {
                    component.setComponentClassPath(XMLUtil.getTextContents(child.getChildNodes()));
                } else if (JBIDescriptorBuilder.BOOTSTRAP_CLASS_NAME_ELEM_NAME.equals(child
                        .getLocalName())) {
                    component.setBootstrapClassName(XMLUtil.getTextContent(child));
                } else if (JBIDescriptorBuilder.BOOTSTRAP_CLASS_PATH_ELEM_NAME.equals(child
                        .getLocalName())) {
                    component.setBootstrapClassPath(XMLUtil.getTextContents(child.getChildNodes()));
                } else if (JBIDescriptorBuilder.SHARED_LIBRARY_ELEM_NAME.equals(child
                        .getLocalName())) {
                    this.loadSharedLibraryListXml(slList, child);
                }
            } else {
                this.loadExtensions(extensions, child);
            }

        }
        // set extensions
        component.setExtensions(extensions);
        // set shared libs
        component.setSharedLibraryList(slList);

        return component;
    }

    /**
     * @param connectionsTree
     * @return
     * @throws JBIDescriptorException
     */
    private List<Connection> loadConnectionsXml(final Node connectionsTree)
            throws JBIDescriptorException {

        final List<Connection> connections = new ArrayList<Connection>();

        if (connectionsTree != null) {
            // get "connections/connection" Node List
            final List<Node> children = XMLUtil.getNodeChildren(connectionsTree);
            for (final Node child : children) {
                if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                    if ("connection".equals(child.getLocalName())) {
                        connections.add(this.loadConnectionXml(child));
                    }
                }
            }
        }
        return connections;
    }

    /**
     *
     * @param node
     * @return
     * @throws JBIDescriptorException
     */
    private Connection loadConnectionXml(final Node node) throws JBIDescriptorException {

        // connection object
        final Connection newItem = new Connection();

        final List<Node> children = XMLUtil.getNodeChildren(node);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if ("consumer".equals(child.getLocalName())) {
                    /*
                     * TODO deal with cases where there is only the interface
                     * name and no endpoint name and service name
                     * newItem.setConsumerInterfaceName(extractXmlAttributeQName(consumer,
                     * "interface-name"));
                     */

                    newItem.setConsumerServiceName(XMLUtil.extractXmlAttributeQName(child,
                            JBIDescriptorBuilder.SERVICE_ATTR_NAME));

                    newItem.setConsumerEndpointName(XMLUtil.getAttributeValue(child,
                            JBIDescriptorBuilder.ENDPOINT_ATTR_NAME));
                } else if ("provider".equals(child.getLocalName())) {
                    try {
                        newItem.setProviderServiceName(XMLUtil.extractRequiredXmlAttributeQName(
                                child, JBIDescriptorBuilder.SERVICE_ATTR_NAME));
                        newItem.setProviderEndpointName(XMLUtil.getRequiredAttributeValue(child,
                                JBIDescriptorBuilder.ENDPOINT_ATTR_NAME));
                    } catch (final NullPointerException e) {
                        throw new JBIDescriptorException(e);
                    }

                }
            }
        }
        return newItem;
    }

    /**
     * @param list
     * @return
     * @throws JBIDescriptorException
     */
    private Consumes loadConsumesXml(final Node node) throws JBIDescriptorException {
        final Consumes newItem = new Consumes();

        // consumes attributes
        try {
            newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                    JBIDescriptorBuilder.INTERFACE_ATTR_NAME));
        } catch (final NullPointerException e) {
            throw new JBIDescriptorException(e);
        }

        newItem.setServiceName(XMLUtil.extractXmlAttributeQName(node,
                JBIDescriptorBuilder.SERVICE_ATTR_NAME));

        newItem.setEndpointName(XMLUtil.getAttributeValue(node,
                JBIDescriptorBuilder.ENDPOINT_ATTR_NAME));

        newItem.setLinkType(XMLUtil.getAttributeValue(node, "link-type"));

        // consumes extensions
        final Extensions extensions = new Extensions();
        final List<Node> children = XMLUtil.getNodeChildren(node);
        for (final Node child : children) {
            this.loadExtensions(extensions, child);
        }
        newItem.setExtensions(extensions);

        return newItem;
    }

    /**
     *
     * @param identificationNode
     * @return
     * @throws JBIDescriptorException
     */
    private Identification loadIdentificationXml(final Node identificationNode)
            throws JBIDescriptorException {
        final Identification identification = new Identification();

        final Extensions extensions = new Extensions();
        // identification sub nodes
        final List<Node> children = XMLUtil.getNodeChildren(identificationNode);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if ("name".equals(child.getLocalName())) {
                    identification.setName(XMLUtil.getTextContent(child));
                } else if ("description".equals(child.getLocalName())) {
                    identification.setDescription(XMLUtil.getTextContent(child));
                }
            } else {
                this.loadExtensions(extensions, child);
            }
        }
        // set extensions
        identification.setExtensions(extensions);

        return identification;
    }

    /**
     * @param list
     * @return
     * @throws JBIDescriptorException
     */
    private Provides loadProvidesXml(final Node node) throws JBIDescriptorException {
        final Provides newItem = new Provides();

        // Provides attributes
        try {
            newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                    JBIDescriptorBuilder.INTERFACE_ATTR_NAME));

            newItem.setServiceName(XMLUtil.extractRequiredXmlAttributeQName(node,
                    JBIDescriptorBuilder.SERVICE_ATTR_NAME));

            newItem.setEndpointName(XMLUtil.getRequiredAttributeValue(node,
                    JBIDescriptorBuilder.ENDPOINT_ATTR_NAME));
        } catch (final NullPointerException e) {
            throw new JBIDescriptorException(e);
        }
        // provides extensions
        final Extensions extensions = new Extensions();
        final List<Node> children = XMLUtil.getNodeChildren(node);
        for (final Node child : children) {
            this.loadExtensions(extensions, child);
        }
        newItem.setExtensions(extensions);

        return newItem;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private ServiceAssembly loadServiceAssemblyXml(final Node node) throws JBIDescriptorException {
        final ServiceAssembly serviceAssembly = new ServiceAssembly();

        final Extensions extensions = new Extensions();
        // Services assembly sub nodes
        final List<Node> children = XMLUtil.getNodeChildren(node);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if (JBIDescriptorBuilder.IDENTIFICATION_ELEM_NAME.equals(child.getLocalName())) {
                    serviceAssembly.setIdentification(this.loadIdentificationXml(child));
                } else if ("service-unit".equals(child.getLocalName())) {
                    serviceAssembly.addServiceUnit(this.loadServiceUnitXml(child));
                } else if ("connections".equals(child.getLocalName())) {
                    serviceAssembly.setConnections(this.loadConnectionsXml(child));
                }
            } else {
                this.loadExtensions(extensions, child);
            }
        }
        // set extensions
        serviceAssembly.setExtensions(extensions);

        return serviceAssembly;
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private Services loadServicesXml(final Node production) throws JBIDescriptorException {

        final Services services = new Services();

        // services attributes
        // is binding component (relative to service unit)
        String isBindingComponent;
        try {
            isBindingComponent = XMLUtil.getRequiredAttributeValue(production, "binding-component");
        } catch (final NullPointerException e) {
            throw new JBIDescriptorException(e);
        }
        services.setBindingComponent(Boolean.parseBoolean(isBindingComponent));

        final Extensions extensions = new Extensions();
        // Services sub nodes
        final List<Node> children = XMLUtil.getNodeChildren(production);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if ("provides".equals(child.getLocalName())) {
                    services.addProvides(this.loadProvidesXml(child));
                } else if ("consumes".equals(child.getLocalName())) {
                    services.addConsumes(this.loadConsumesXml(child));
                }
            } else {
                this.loadExtensions(extensions, child);
            }
        }
        // set extensions
        services.setExtensions(extensions);

        return services;
    }

    /**
     * @param list
     * @return
     * @throws JBIDescriptorException
     */
    private ServiceUnit loadServiceUnitXml(final Node node) throws JBIDescriptorException {

        final ServiceUnit result = new ServiceUnit();

        final Extensions extensions = new Extensions();
        // Service unit sub nodes
        final List<Node> children = XMLUtil.getNodeChildren(node);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if (JBIDescriptorBuilder.IDENTIFICATION_ELEM_NAME.equals(child.getLocalName())) {
                    result.setIdentification(this.loadIdentificationXml(child));
                } else if ("target".equals(child.getLocalName())) {
                    final List<Node> childChildren = XMLUtil.getNodeChildren(child);
                    for (final Node targetChild : childChildren) {
                        if ("artifacts-zip".equals(targetChild.getLocalName())) {
                            result.setTargetArtifactsZip(XMLUtil.getTextContent(targetChild));
                        } else if ("component-name".equals(targetChild.getLocalName())) {
                            result.setTargetComponentName(XMLUtil.getTextContent(targetChild));
                        }
                    }
                }
            } else {
                this.loadExtensions(extensions, child);
            }
        }
        // set extensions
        result.setExtensions(extensions);

        return result;
    }

    /**
     * @param list
     * @return
     */
    private void loadSharedLibraryListXml(final List<SharedLibraryList> slList, final Node node) {
        final SharedLibraryList sharedLibrary = new SharedLibraryList();
        sharedLibrary.setName(XMLUtil.getTextContent(node));
        sharedLibrary.setVersion(XMLUtil.getAttributeValue(node,
                JBIDescriptorBuilder.VERSION_ATTR_NAME));
        slList.add(sharedLibrary);
    }

    /**
     * @param element
     * @throws JBIDescriptorException
     */
    private SharedLibrary loadSharedLibraryXml(final Node production) throws JBIDescriptorException {
        final SharedLibrary sharedLibrary = new SharedLibrary();

        // shared library attributes
        sharedLibrary.setClassLoaderDelegation(XMLUtil.getAttributeValue(production,
                "class-loader-delegation"));
        sharedLibrary.setVersion(XMLUtil.getAttributeValue(production,
                JBIDescriptorBuilder.VERSION_ATTR_NAME));

        final List<Node> children = XMLUtil.getNodeChildren(production);
        for (final Node child : children) {
            if (child.getNamespaceURI().equals(JBIDescriptorBuilder.NAMESPACE_JBI)) {
                if (JBIDescriptorBuilder.IDENTIFICATION_ELEM_NAME.equals(child.getLocalName())) {
                    sharedLibrary.setIdentification(this.loadIdentificationXml(child));
                } else if ("shared-library-class-path".equals(child.getLocalName())) {
                    sharedLibrary.setSharedLibraryClassPath(XMLUtil.getTextContents(child
                            .getChildNodes()));
                }
            }
        }
        return sharedLibrary;
    }

    /**
     * Get the Petals Schema
     * @return the petals schema
     * @throws JBIDescriptorException
     *     Impossible to get the PEtALS schema
     */
    private Schema getPetalsSchema() throws JBIDescriptorException {
        Schema schema = null;
        final SchemaFactory factory = SchemaFactory.newInstance(JBIDescriptorBuilder.NAMESPACE_XSD);

        final Source[] schemas = new StreamSource[2];

        try {
            schemas[0] = new StreamSource(this
                    .openSchemaFile(JBIDescriptorBuilder.PETALS_EXTENSIONS_XSD));
            schemas[1] = new StreamSource(this.openSchemaFile(JBIDescriptorBuilder.PETALS_XSD));

            schema = factory.newSchema(schemas);

        } catch (final IOException e) {
            throw new JBIDescriptorException("IOException: Cannot load the Petals Schema", e);
        } catch (final SAXException e) {
            throw new JBIDescriptorException("SAXException: Cannot load the Petals Schema", e);
        }
        return schema;
    }

    /**
     * Get the JBI Schema
     * @return the JBI schema
     * @throws JBIDescriptorException
     *     Impossible to get the JBI schema
     */
    private Schema getJBISchema() throws JBIDescriptorException {
        Schema schema = null;
        final SchemaFactory factory = SchemaFactory.newInstance(JBIDescriptorBuilder.NAMESPACE_XSD);

        final Source[] schemas = new StreamSource[1];

        try {
            schemas[0] = new StreamSource(this.openSchemaFile(JBIDescriptorBuilder.JBI_XSD));

            schema = factory.newSchema(schemas);

        } catch (final IOException e) {
            throw new JBIDescriptorException("IOException: Cannot load the JBI Schema", e);
        } catch (final SAXException e) {
            throw new JBIDescriptorException("SAXException: Cannot load the JBI Schema", e);
        }
        return schema;
    }

    /**
     * NOTE: The first location to search for JBI schema files is within the
     * Petals installation root directory. If a schema file is not found within
     * root directory, the alternative is use the packaged schema files flatten
     * somewhere which can be found in class path. It won't work if the schema
     * are compressed in a Zip or Jar archive.
     *
     * @return JBI schema as InputStream
     * @throws IOException
     *             if an error occurs during InputStream creation from schema
     *             file
     */
    private InputStream openSchemaFile(final String schemaFile) throws IOException {
        InputStream theSchemaResource = null;
        // search through classpath obtains from this class classloader
        final URL schemaURL = this.getClass().getResource("/" + schemaFile);

        if (schemaURL != null) {
            theSchemaResource = schemaURL.openStream();
        }

        if (theSchemaResource == null) {
            // search in petals installation root
            String petalsRoot = System.getenv(JBIDescriptorBuilder.PETALS_HOME_PROPERTY);
            if (petalsRoot == null) {
                petalsRoot = System.getProperty(JBIDescriptorBuilder.PETALS_HOME_PROPERTY);
            }
            if (petalsRoot != null) {
                theSchemaResource = new FileInputStream(new File(petalsRoot
                        + JBIDescriptorBuilder.SCHEMA_RESOURCE + File.separator + schemaFile));
            }
        }
        return theSchemaResource;
    }

}
